/*!
 * Copyright (c) Jayly. All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */
import { ScoreboardIdentityType, system, world } from "@minecraft/server";
const version = "1.1.0";
const str = () => ('00000000000000000' + (Math.random() * 0xffffffffffffffff).toString(16)).slice(-16);
const uuid = () => {
    const [a, b] = [str(), str()];
    return `${a.slice(0, 8)}-${a.slice(8, 12)}-4${a.slice(13)}-a${b.slice(1, 4)}-${b.slice(4)}`;
};
const allowedTypes = ["string", "number", "boolean"];
const encrypt = (data, salt) => {
    const encryptedChars = [];
    for (let i = 0; i < data.length; i++) {
        const charCode = data.charCodeAt(i) + salt.charCodeAt(i % salt.length);
        encryptedChars.push(charCode);
    }
    return String.fromCharCode(...encryptedChars);
};
const decrypt = (encrypted, salt) => {
    const decryptedChars = [];
    for (let i = 0; i < encrypted.length; i++) {
        const charCode = encrypted.charCodeAt(i) - salt.charCodeAt(i % salt.length);
        decryptedChars.push(charCode);
    }
    return String.fromCharCode(...decryptedChars);
};
const CreateCrashReport = (action, data, error, salt) => {
    console.warn("[JaylyDB] Failed to " + action + " JSON data.", "\nVersion: " + version, "\nData: " + data, "\nSalt: " + salt, "\nError: " + error.message, "\n" + error.stack);
    throw new Error(`Failed to ${action} data. Please check content log file for more info.\n`);
};
const DisplayName = {
    parse(text, salt) {
        try {
            const a = salt ? decrypt(text, salt) : text;
            return JSON.parse(`{${a}}`);
        }
        catch (error) {
            CreateCrashReport("load", text, error, salt);
        }
    },
    stringify(value, salt) {
        try {
            const a = JSON.stringify(value).slice(1, -1);
            return salt ? encrypt(a, salt) : a;
        }
        catch (error) {
            CreateCrashReport("save", JSON.stringify(value), error, salt);
        }
    }
};
;
class JaylyDB {
    objective;
    encrypted;
    localState = new Map();
    salt;
    SYNC_OK = true;
    updateParticipants() {
        this.localState.clear();
        for (let participant of this.objective.getParticipants()) {
            if (participant.type !== ScoreboardIdentityType.FakePlayer)
                continue;
            const data = DisplayName.parse(participant.displayName.replace(/\\"/g, `"`), this.salt);
            const key = Object.keys(data)[0];
            const value = data[key];
            this.localState.set(key, {
                identity: participant,
                encoded_value: participant.displayName.replace(/\\"/g, `"`),
                decoded_value: value
            });
            if (participant.displayName.includes(`\\"`)) {
                const encoded = DisplayName.stringify({ [key]: value }, this.salt);
                this.objective.removeParticipant(participant);
                this.objective.setScore(encoded, 0);
            }
        }
        ;
    }
    constructor(id, encrypted = false) {
        this.objective = world.scoreboard.getObjective("jaylydb:" + id) ?? world.scoreboard.addObjective("jaylydb:" + id, uuid());
        this.encrypted = encrypted;
        this.salt = this.encrypted ? this.objective.displayName : undefined;
        this.updateParticipants();
        system.runInterval(() => {
            const objective = world.scoreboard.getObjective("jaylydb:" + id);
            if (objective) {
                if (this.SYNC_OK === false) {
                    for (const [, value] of this.localState.entries())
                        this.objective.setScore(value.encoded_value, 0);
                    console.log(`[JaylyDB] Database '${objective.id.slice(6)}' is now synced.`);
                }
                else
                    this.updateParticipants();
                this.SYNC_OK = true;
            }
            else if (this.SYNC_OK === true) {
                console.error(`[JaylyDB] There is a sync issue with database '${id}'.`);
                this.SYNC_OK = false;
            }
        });
    }
    get size() {
        return this.localState.size;
    }
    clear() {
        this.localState.forEach(participant => this.objective.removeParticipant(participant.identity));
        this.localState.clear();
    }
    delete(key) {
        const participant = this.localState.get(key);
        if (!participant)
            return false;
        const success = this.objective.removeParticipant(participant.identity);
        this.localState.delete(key);
        return success;
    }
    forEach(callbackfn) {
        for (const [key, value] of this.entries())
            callbackfn(value, key, this);
    }
    get(key) {
        if (!this.localState.has(key))
            this.updateParticipants();
        return this.localState.get(key)?.decoded_value;
    }
    has(key) {
        return this.localState.has(key);
    }
    set(key, value) {
        if (!allowedTypes.includes(typeof value))
            throw new TypeError("JaylyDB::set only accepts a value of string, number, or boolean.");
        if (this.localState.get(key)?.decoded_value === value)
            return this;
        const encoded = DisplayName.stringify({ [key]: value }, this.salt);
        if (encoded.length > 32767)
            throw new RangeError("JaylyDB::set only accepts a string value less than 32767 characters.");
        if (this.localState.has(key))
            this.objective.removeParticipant(this.localState.get(key).identity);
        this.objective.setScore(encoded, 0);
        const data = {
            encoded_value: encoded,
            decoded_value: value,
            identity: this.objective.getParticipants().find(participant => participant.displayName === encoded),
        };
        this.localState.set(key, data);
        return this;
    }
    *entries() {
        for (const [key, data] of this.localState.entries())
            yield [key, data.decoded_value];
    }
    *keys() {
        for (const [key] of this.entries())
            yield key;
    }
    *values() {
        for (const [, value] of this.entries())
            yield value;
    }
    [Symbol.iterator]() {
        return this.entries();
    }
    [Symbol.toStringTag] = JaylyDB.name;
}
;
export { JaylyDB };
